import cupy as cp
import numpy as np
import os
import math
import random
import re
import sys

cp.random.seed()
cp.set_printoptions(precision=2,floatmode='fixed',suppress=True)

class Distributions(object):

  """Random training distributions for test modules."""

  def __init__(self):
    pass

  def linear_distribution(self, depth):
    """linear regression that predicts y from x for x-values on a random line with slope and intercept

        Args:
          depth (int):
              The number of x-values to generate.

        Returns:
          tuple of (generated vector of x-values, vector of expected y-values)
    """
    # random input values
    x = cp.random.random((depth,1)).astype(cp.float32)
    # target weight, bias
    z = cp.random.random((2,1)).astype(cp.float32)
    # y = wx + b
    y = [ z[0] * x[j] + z[1] for j in range(0,len(x)) ]
    y = cp.array(y).astype(cp.float32)
    return x,y

  def logical_distribution(self, depth, mode):
    """boolean logic

        Args:
          depth (int):
              The number of rows for the generated two-column binary matrix.
          mode (str):
              Accepts "and", "or", "xor", or "xnor".

        Returns:
          tuple of (generated binary matrix, expected output)
    """
    x = cp.round(cp.random.random((depth,2))).astype(cp.float32)
    y = cp.zeros([depth,1]).astype(cp.float32).flatten()
    if mode=="and":
      y = cp.array([x[j,0] * x[j,1] for j in range(0,len(x))])
    elif mode=="or":
      y = cp.array( [ cp.max(x[j]) for j in range(0,len(x)) ] )
    elif mode=="xor":
      y = cp.array([abs(x[j,0] - x[j,1]) for j in range(0,len(x))])
    elif mode=="xnor":
      y = cp.array([1 if x[j,0]==x[j,1] else 0 for j in range(0,len(x))])
    else:
      return None,None
    return x,y

  def arithmetic_distribution(self, depth, mode):
    """predict arithmetic result from distributions of two input values

        Args:
          depth (int):
              The number of rows for the generated matrix of floating point numbers.
          mode (str):
              The mode of operation. Accepts either "add", "subtract", "multiply", "divide", or "zero" (always predict 0).

        Returns:
          tuple of (generated matrix, expected output)

        Raises:
          SystemExit
              If the provided `mode` is not "summation" or "one_hot".
    """
    x = cp.random.random((depth,2)).astype(cp.float32)
    y = cp.zeros([depth,1]).astype(cp.float32)
    if mode=="add":
      y = cp.array([x[j,0] + x[j,1] for j in range(0,len(x))]) # weights = 1
    elif mode=="subtract":
      y = cp.array([x[j,0] - x[j,1] for j in range(0,len(x))]) # w1 = 1, w2 = -1
    elif mode=="multiply":
      y = cp.array([x[j,0] * x[j,1] for j in range(0,len(x))])
    elif mode=="divide":
      y = cp.array([x[j,0] / x[j,1] for j in range(0,len(x))])
    elif mode=="zero": # weights = 0
      y = y.flatten()
    else:
      return None,None
    return x,y

  def summation(self, rows, cols, mode="one_hot"):
    """distributions of binary vectors for testing binary cross entropy (one-hot mode only)

        Args:
          rows (int):
              The number of rows for the generated binary matrix.
          cols (int):
              The number of columns for the generated binary matrix.
          mode (str):
              The mode of operation. Accepts either "summation" or "one_hot".
              - "summation": Produces a scalar count of the number of ones in each x vector.
              - "one_hot": Produces a one-hot encoded representation of the count of ones in each x vector.
              Defaults to "one_hot".

        Returns:
          tuple of (generated binary matrix, expected output)

        Raises:
          SystemExit if the provided `mode` is not "summation" or "one_hot".
    """
    x = cp.array([random.randrange(2) for i in range(rows * cols)]).reshape(rows,cols)
    if mode=="summation":
      # y = a scalar for the number of ones in each x vector
      y = cp.array([[cp.sum(x[i])] for i in range(0,len(x))])
    elif mode=="one_hot":
      # y = one hot vector with a 1 in the place of the number of ones in x vector
      y = cp.array([[0] * (cols + 1) for i in range(rows)])
      for i in range(rows):
        y[i,cp.sum(x[i])] = 1
    else:
      sys.exit("mode must be summation or one_hot")
    return x,y

  def sort(self, rows, cols):
    """numerical sort

        Args:
          rows (int):
            the number of rows for the generated matrix
          cols (int):
            the number of columns for the generated matrix

        Returns:
          tuple of (generated matrix, sorted matrix)
    """
    x = cp.random.random((rows,cols))
    y = cp.array([sorted(x[i]) for i in range(0,len(x))])
    return x,y



from nillanet.model import NN
from nillanet.activations import Activations
from nillanet.loss import Loss
from nillanet.distributions import Distributions
from nillanet.initializer import Initializer

d = Distributions()
x,y = d.arithmetic_distribution(10,"subtract")
print(x.shape)
print(y.shape)

a = Activations()
activation = a.tanh
derivative1 = a.tanh_derivative
resolver = a.linear
derivative2 = a.linear_derivative

l = Loss()
loss = l.mse
derivative3 = l.mse_derivative

input = x
output = y
features = x.shape[1]
architecture = [2,4,1]
learning_rate = 0.01
epochs = 1000
initializer = Initializer(distribution=Initializer.xavier, low=-2, high=2, mean=0, std=1)

model = NN(features,architecture,activation,derivative1,resolver,derivative2,loss,derivative3,learning_rate,initializer=initializer)
model.summary()

for epoch in range(epochs):
    model.train(input,output,epoch,epochs,verbose=True,step=100,autosave=True)

prediction = model.predict(input)

print("prediction")
print(prediction)
print("expected")
print(y)

from nillanet.io import IO
io = IO()
best = io.load(model.backup)
prediction = best.predict(x)
print("best")
print(prediction)


import math

class Scheduler:

    """Learning rate scheduler for NN class"""

    def __init__(self, mode, lr, lowbound=1e-8, scaler=0, warmup=0, interval=1, maxsteps=0, custom=None):

        """Read parameters for initializing the scheduler.

        Args:
            mode (str):
                The mode of learning rate decay.
                Required.
            lr (float):
                The initial learning rate.
                Required.
            lowbound (float):
                The lower bound for the learning rate.
                Default: 1e-8.
            scaler (float):
                Mode:
                    The scaling factor for the constant mode only.
                Range:
                    { x | 0 < x < 1 }.
                Optional:
                    Set zero to skip.
            warmup (int):
                The number of epochs for an optional warmup period.
                Optional, set zero to skip.
            interval (int):
                The interval at which a step is applied.
                Default: 1.
            maxsteps (int):
                The maximum number of updates applied to the learning rate.
                Optional, set zero to skip.
            custom (function):
                A custom function for updating the learning rate.
                Optional, set None to skip.

        Attributes:
            sigma (float): the current learning rate
            steps (int): the number of updates applied so far
        """

        self.mode = mode
        if self.mode not in ['inverse', 'linear', 'cosine', 'constant', 'custom']:
            raise ValueError(f"unknown mode {self.mode}")

        # remains constant throughout training
        self.lr = lr
        # varies as training progresses
        self.sigma = self.lr
        # lowbound for lr
        self.lowbound = lowbound

        # scaling factor for constant mode e.g. 0.90
        self.scaler = scaler

        # units: epochs, defines a warmup period of reduced lr, set zero to skip
        self.warmup = warmup
        # units: epochs, interval at which a step is applied
        self.interval = interval

        # progress count with coarser granularity than epochs
        self.steps = 0
        # units: steps
        self.maxsteps = maxsteps

        # optional custom function for updating lr
        # must set mode = 'custom'
        self.custom = custom

    def step(self, epoch, epochs):
        """Update the learning rate based on the current epoch and mode.

        Args:
            epoch (int): the current epoch
            epochs (int): the total number of epochs

        Returns:
            sigma (float): the updated learning rate
        """

        if self.warmup > 0 and epoch <= self.warmup:
            epoch = epoch + 1
            self.sigma = self.lr * (epoch / self.warmup)
            return self.sigma
        elif epoch % self.interval != 0:
            return self.sigma
        elif self.maxsteps > 0 and self.steps > self.maxsteps:
            return self.sigma

        self.steps += 1

        if self.mode == 'inverse':
            self.sigma = self.inverse(epoch, epochs)
        elif self.mode == 'linear':
            self.sigma = self.linear(epoch, epochs)
        elif self.mode == 'cosine':
            self.sigma = self.cosine(epoch, epochs)
        elif self.mode == 'constant':
            if self.scaler <= 0 or self.scaler >= 1:
                raise ValueError("scaler must be between zero and one")
            self.sigma = self.constant(epoch, epochs)
        elif self.mode == 'custom':
            if self.custom is None:
                raise ValueError("custom function must be provided")
            self.sigma = self.custom(epoch, epochs)

        return max(self.sigma, self.lowbound)

    def inverse(self, epoch, epochs):
        """varies sigma as the inverse square of the number of steps taken"""
        self.sigma *= (self.steps ** -0.5)
        return self.sigma

    def linear(self, epoch, epochs):
        """varies sigma linearly as the number of remaining epochs"""
        self.sigma *= (1 - epoch / epochs)
        return self.sigma

    def cosine(self, epoch, epochs):
        """varies sigma trigonometrically from lr to lowbound"""
        self.sigma = self.lowbound + (abs(self.lr - self.lowbound) / 2) * (1 + math.cos(math.pi * epoch / epochs))
        return self.sigma

    def constant(self, epoch, epochs):
        """varies sigma by a constant factor of scaler which ranges from 0 to 1"""
        self.sigma *= self.scaler
        return self.sigma



import math
import scipy

class Scheduler:
    def __init__(self, lr, mode, interval=1, lowbound=1e-8, maxsteps=0, warmup=0):
        self.lr = lr
        self.delta = self.lr
        self.mode = mode
        self.interval = interval # epochs interval at which update applied
        self.lowbound = lowbound # lowbound for lr
        self.steps = 0
        self.maxsteps = maxsteps # units: steps
        self.warmup = warmup # units: epochs
    def step(self, epoch, epochs):
        if epoch % self.interval != 0:
            return self.delta
        self.steps += 1
        if self.maxsteps > 0 and self.steps > self.maxsteps:
            return self.delta
        if self.warmup > 0 and epoch < self.warmup:
            epoch = epoch + 1
            self.delta = self.lr * (epoch / self.warmup)
        if self.mode == 'inverse':
            self.delta = self.delta * (self.steps ** -0.05)
        elif self.mode == 'cosine':
            self.delta = self.lowbound + (abs(self.lr - self.lowbound) / 2) * (1 + math.cos(math.pi * epoch / epochs))
        elif self.mode == 'exponential':
            self.delta = self.delta * math.exp(-0.01 * epoch)
        elif self.mode == 'linear':
            self.delta = self.delta * (1 - epoch / epochs)
        elif self.mode == 'constant':
            self.delta = self.lr
        else:
            raise ValueError(f"mode {self.mode} not recognized")
        return max(self.delta, self.lowbound)



from scipy import stats
import sys

class Initializer(object):

    """Weight distributions for custom initializations."""

    def __init__(self, distribution=None, low=0.0, high=1.0, mean=0.0, std=1.0):
        """Models a configurable statistical distribution for initializing weights.

        Args:
            distribution (function): A function representing the desired
                distribution. If None, a default normal distribution will be used.
            low (float):
                normal distribution:
                    the lower boundary for standard deviations away from the mean
                otherwise:
                    the lower boundary of the abscissae of the distribution
                Defaults to 0.0.
            high (float):
                normal distribution:
                    the upper boundary for standard deviations away from the mean
                otherwise:
                    the upper boundary of the abscissae of the distribution
                Defaults to 1.0.
            mean (float): The mean value for the distribution. Defaults to 0.0.
            std (float): The standard deviation for the distribution. Defaults to 1.0.
        """
        self.distribution = distribution
        if distribution is None:
            self.distribution = self.normal
        self.low = low
        self.high = high
        self.mean = mean
        self.std = std

    def __call__(self, shape):
        return self.distribution(self, shape)

    def normal(self, shape):
        """Generates random numbers from a bell-shaped distribution within the defined range.

        Note that the `low` and `high` parameters are interpreted as standard deviations away from the mean.

        Args:
            shape (tuple): The desired shape of the output array containing samples.

        Returns:
            numpy.ndarray
                An array of random samples drawn from the truncated normal distribution.
        """
        return stats.truncnorm.rvs(self.low, self.high, loc=self.mean, scale=self.std, size=shape)

    def uniform(self, shape):
        """Generates random numbers from a plateau-shaped distribution within the defined range.

        The `low` and `high` parameters are interpreted as the lower and upper bounds of the plateau.

        Args:
            shape (tuple): The desired shape of the output array containing samples.

        Returns:
            numpy.ndarray
                An array of random samples drawn from the uniform distribution.
        """
        return stats.uniform.rvs(loc=self.low, scale=self.high-self.low, size=shape)

    def xavier(self, shape):
        """Generates random numbers with variance equal to 1 over the square of n.

        `Xavier` demonstrated its usefulness for tanh or sigmoid activation functions.
        The `low` and `high` parameters are interpreted as the lower and upper bounds of the plateau.

        Args:
            shape (tuple): The desired shape of the output array containing samples.

        Returns:
            numpy.ndarray
                An array of random samples drawn from the truncated normal distribution.
        """
        return stats.truncnorm.rvs(self.low, self.high, scale=1.0/shape[0]**0.5, size=shape)

    def he(self, shape):
        """Generates random numbers with variance equal to 2 over the square of n.

        `He` demonstrated its usefulness for the relu activation function.
        The `low` and `high` parameters are interpreted as the lower and upper bounds of the plateau.

        Args:
            shape (tuple): The desired shape of the output array containing samples.

        Returns:
            numpy.ndarray
                An array of random samples drawn from the truncated normal distribution.
        """
        return stats.truncnorm.rvs(self.low, self.high, scale=2.0/shape[0]**0.5, size=shape)

if __name__ == '__main__':
    shape = (3, 3)
    low,high,mean,std = (
        0.0, 1.0, 0.0, 1.0
    )
    if len(sys.argv) >= 3:
        low = float(sys.argv[1])
        high = float(sys.argv[2])
    if len(sys.argv) >= 5:
        mean = float(sys.argv[3])
        std = float(sys.argv[4])
    print("shape %s low %s high %s mean %s std %s" % (shape, low, high, mean, std))
    print("normal")
    intl = Initializer(distribution=Initializer.normal, low=low, high=high, mean=mean, std=std)
    print(intl.normal(shape))
    print("uniform")
    intl = Initializer(distribution=Initializer.uniform, low=low, high=high, mean=mean, std=std)
    print(intl.uniform(shape))
    print("xavier")
    intl = Initializer(distribution=Initializer.xavier, low=low, high=high, mean=mean, std=std)
    print(intl.xavier(shape))
    print("he")
    intl = Initializer(distribution=Initializer.he, low=low, high=high, mean=mean, std=std)
    print(intl.he(shape))
